/* vi: set sw=4 ts=4: */
/*
 * KON2 - Kanji ON Console -
 * Copyright (C) 1992-1996 Takashi MANABE (manabe@papilio.tutics.tut.ac.jp)
 *
 * CCE - Console Chinese Environment -
 * Copyright (C) 1998-2003 Rui He (rhe@3eti.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

/* defs.h -- global definition */
#ifndef __CCE_DEFS_H__
#define __CCE_DEFS_H__

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osdep.h"

#undef FALSE
#undef TRUE

#if defined(__BEOS__)
#define bool CCE_BOOL    /* bool is already defined in BeOS */
#endif

typedef	enum {FALSE, TRUE} bool;

#define	FAILURE	(-1)
#define SUCCESS	(0)

#define CCE_VERSION_STR "Version " VERSION " (Jan 2004)"

#define CCE_PATHNAME_MAX  255

static inline void bzero2(void *head, int n)
{
	while (n-- > 0)
		*((unsigned char *)head)++ = (unsigned char)0x0;
}

static inline void wzero(void *head, int n)
{
	n >>= 1;
	while (n-- > 0)
		*((unsigned short *)head)++ = (unsigned short)0x0;

}

static inline void lzero(void *head, int n)
{
	n >>= 2;
    while (n-- > 0)
        *((unsigned int *)head)++ = (unsigned int)0x0;
}

static inline void bmove(void *dst, void *src, int n)
{
	while(n-- > 0)
		*((unsigned char *)dst)++ = *((unsigned char *)src)++;
}

static inline void brmove(void *dst, void *src, int n)
{
    while(n-- > 0)
        *--((unsigned char *)dst) = *--((unsigned char *)src);
}

static inline void wmove(void *dst, void *src, int n)
{
	n >>= 1;
    while(n-- > 0)
        *((unsigned short *)dst)++ = *((unsigned short *)src)++;
}

static inline void lmove(void *dst, void *src, int n)
{
	n >>= 2;
	while(n-- > 0)
		*((unsigned int *)dst)++ = *((unsigned int *)src)++;
}

static inline void SafeFree(void **p)
{
        if (*p) 
        {
                free(*p);
                *p = NULL;
        }
}

#endif  /* __CCE_DEFS_H__ */
