/* vi: set sw=4 ts=4: */
/*
 * CCE - Console Chinese Environment -
 * Copyright (C) 1998-2003 Rui He (rhe@3eti.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>

#include "config.h"
#include "errors.h"
#include "encfilter.h"

char *IconvSrcEncoding[][4] =
{ 
  {"ISO-2022-JP",},
  {"EUC-JP",},
  {"SJIS",},
  {"BIG5",},
  {"GB2312",},
  {"GBK",},  /* CP936 */
  {"CP949",},  /* CP949 is UHC, like GBK space, but easier to use iconv? */
  {"UHC",},  /* CP949 */
  {"UTF8",}, /* or UTF16? */
  {NULL},
};
 
#ifdef SUPPORT_ENCODING_FILTER

 int UseEncodingFilter;  /* use filter such as UTF-8 */
 char *EncodingFilter;   /* filter's name */
 char EncodingFilterBuff[BUFSIZ];   /* global buffer used by ConsoleHandler() and TextCopy() */
 size_t EncodingFilterLen; 

 static iconv_t FilterCD[2]; /* 0 is from filter to target, 1 is reverse */
 
void CleanupEncodingFilter(void)
{
	if (FilterCD[0] && FilterCD[0] != (iconv_t)-1) iconv_close(FilterCD[0]);
	if (FilterCD[1] && FilterCD[1] != (iconv_t)-1) iconv_close(FilterCD[1]);
	FilterCD[0] = FilterCD[1] = NULL;
}

int SetupEncodingFilter(int sysCoding)
{
	if (UseEncodingFilter == 0) return 0; /* success */
	if (!EncodingFilter || !IconvSrcEncoding[sysCoding][0]) return -1;

	CleanupEncodingFilter(); /* close old ones */

	UseEncodingFilter = 0;
	if ((FilterCD[0] = iconv_open(IconvSrcEncoding[sysCoding][0], EncodingFilter)) == (iconv_t)-1)
	{
		error("Failed to do iconv_open(%s,%s), filter disabled.\r\n", 
			IconvSrcEncoding[sysCoding][0], EncodingFilter);
		return -1;
	}
	
	if ((FilterCD[1] = iconv_open(EncodingFilter, IconvSrcEncoding[sysCoding][0])) == (iconv_t)-1)
	{
		iconv_close(FilterCD[0]);
		FilterCD[0] = NULL;
		return -1;
	}
	
	message("Using encoding filter %s for %s.\r\n", EncodingFilter, IconvSrcEncoding[sysCoding][0]);

	UseEncodingFilter = 1;
	return 0;
}

int DoEncodingFilter(int dir, char *inbuf, size_t inlen)
{
	size_t ret;
	char *oldinbuf = inbuf;
	char *outbuf = EncodingFilterBuff;

	if (!inbuf || inlen <= 0) return 0;

	EncodingFilterLen = sizeof(EncodingFilterBuff);

	ret = iconv(FilterCD[dir], &inbuf, &inlen, &outbuf, &EncodingFilterLen);

	EncodingFilterLen = sizeof(EncodingFilterBuff) - EncodingFilterLen; /* converted  */

	if (ret == -1)  
	{
		switch(errno)
		{
		case EINVAL:  /* incomplete, inbuf is pointing to incomplete buffer */
		case E2BIG:  /* output room is not enough */
			memcpy(oldinbuf, inbuf, inlen);  /* save them to the beginning */
			return inlen;
			
		case EILSEQ:  /* illegal sequence */
		default:
			return 0; 
		}
	}

	return 0; /* inlen should be 0 since everything is converted */
}

#endif /* SUPPORT_ENCODING_FILTER */

