/* vi: set sw=4 ts=4: */
/*
 * CCE - Console Chinese Environment -
 * Copyright (C) 1998-2003 Rui He (rhe@3eti.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY MAEDA ATUSI AND THE AUTHOR ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>

#include <defs.h>
#include <convert.h>
#include <hzinput.h>

u_char *CurCoding[] = { "ڡ2022-JP","EUC-JP", "ySJISz", "iBig5j", "GB", "GBK" ,"KSC",
		  				"KSC", /* UHC */ "[Unicode]",
				 	  };

u_char *FullCharMap[] = { NULL, /* 0,  */
	"ɡǡʡˡܡ䡩"
	"£ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣ڡ̡͡"
	"ƣСáѡ  ", /* EUC-JP */

	"@Ihfij{C]D^OPQRSTUVWXFGH"
	"`abcdefghijklmnopqrstuvwxyk_lOQ"
	"eobp`  ", /*  SJIS */

	"@ICH]^ϡAСDGQաס֡H"
	"IϢТѢҢӢԢբ֢עآ٢ڢۢܢݢޢߢe@fsZ"/* */
	"@ABCW  ", /* BIG5 */

	"磥"
	"£ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣ڡܡޣ"
	"  ", /* GB */
	
	"磥"
	"£ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣ڡܡޣ"
	"  ", /* GBK */

    ""
    "£ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣ڣۡݣޣ"
    "", /* KSC */

    ""
    "£ãģţƣǣȣɣʣˣ̣ͣΣϣУѣңӣԣգ֣ףأ٣ڣۡݣޣ"
    "", /* UHC */
	
	NULL /* Unicode */
 }; 

	char *InpNameStr[] = { 
							 "%s%s",  /* ISO-JP */
	                         "%s%s",  /* EUC */
							 "y%s%sz",  /* SJIS */
							 "i%s%sj", /* Big5 */
							 "%s%s", /* GB */
							 "%s%s", /* GBK */
							 "%s%s", /* KSC */
							 "%s%s", /* UHC */
							 " [%s%s] ", /* Unicode */
							};

 #if 0 
	char *HalfFullStr[][2] = { 
							 {"Ⱦ%s", "%s"}, /* ISO-JP */
	                         {"Ⱦ%s", "%s"}, /* EUC */
							 {"y%sz", "yS%sz"}, /* SJIS */
							 {"ib%sj","i%sj"}, /* Big5 */
							 {"%s","ȫ%s"}, /* GB */
							 {"%s","ȫ%s"}, /* GBK */
							 {"%s","%s"}, /* KSC */
							 {"%s","%s"}, /* UHC */
							 {"[Half%s]","[Full%s]"}, /* Unicode */
							};

	char *ChinesePunctStr[][2] = {
                                  {",.", ""},  /* ISO */
								  {",.", ""},  /* EUC  GB & JIS same code, 0xA1A3 */
                                  {",.", "B"}, /* SJIS */
								  {",.", "C"}, /* Big5 "A" */
								  {",.", ""},  /* GB  ""  */
                                  {",.", ""}, /* GBK */
                                  {",.", ""}, /* KSC, KSC's period 0xA1A3 upper right? */
                                  {",.", ""}, /* UHC */
                                  {",.", ",."}, /* Unicode */
                                 };
								
#endif

  u_char HalfFullIcons[2][32] = {

  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x20, 0x00, 0x30, 0x00, 0x38, 0x00, 0x38, 0x00, 0x3C, 0x00, 0x7C,
    0x00, 0x7C, 0x01, 0xFC, 0x1F, 0xFC, 0x0F, 0xF8, 0x07, 0xF0, 0x01, 0xE0, 0x00, 0x00, 0x00, 0x00,  },
 
   { 0x00, 0x00, 0x00, 0x00, 0x07, 0xE0, 0x0F, 0xF0, 0x1F, 0xF8, 0x3F, 0xFC, 0x3F, 0xFC, 0x3F, 0xFC,
    0x3F, 0xFC, 0x3F, 0xFC, 0x3F, 0xFC, 0x1F, 0xF8, 0x0F, 0xF0, 0x07, 0xE0, 0x00, 0x00, 0x00, 0x00,  },
 
};

  u_char ChEngPunctIcons[2][32] = {

  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x18, 0x38,
    0x3C, 0x38, 0x18, 0x18, 0x00, 0x10, 0x00, 0x20, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
  },
 { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x3C, 0x1C, 0x66, 0x3E, 0x42, 0x3E, 0x42, 0x3E,
   0x66, 0x1E, 0x3C, 0x0E, 0x00, 0x0C, 0x00, 0x38, 0x00, 0x70, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
 },
};

	char *EngNumStr[] = { 
		"ڱѿ%s%s", "ڱѿ%s%s", "yp%sz%s", 
		"i^%sj%s", /* big5 */
		"Ӣ%s%s", "Ӣ%s%s", /* GB/GBK */
		"%s%s", /* KSC */
		"%s%s", /* UHC */
		"[ENGNUM%s]%s", /* Unicode */
	 };
	
	char *WindowStr[] = {
    	"ڽü%dۡٸ%d", "ڽü%dۡٸ%d", "yI[%dzyx%dz",  /* EUC, SJIS */
		"i׺%djif%dj", /* big 5 */
		"ն%d%d", "ն%d%d", /* GB/GBK */
		"Ӯ%dâ%d", /* KSC no Window Hanzi in Korean?  */
		"Ӯ%dâ%d", /* UHC no Window Hanzi in Korean?  */
		"[Term %d][Window %d]", /* Unicode */
	};

	char *LianXiangStr[] = 
	   { "", "", "z",  /* ISO, EUC, SJIS */
		 "pQ", "", "",  /* Big5 GB GBK */
		 "֤", "֤", /* KSC UHC */
         "Asso",
       };
	
	char *IntCode_CNameStr[] = { "", "", "",
		                        "X", "", "", 
								"Үا",	"Үا", /* KSC UHC */
								"InCode",  /* Unicode */

							  };

	char *Pinyin_CNameStr[] = { "Ҳ","Ҳ", "q",    /* Japanese Kanji has no "Pin" ? so I used Yin :) */
								"", "","",
                                "", "",   /* Korean Hanzi has no Pin too? Used Yin */
							    "Pinyin",  /* Unicode */
							  };

	char *Zhuyin_CNameStr[] = { "첻","첻", "",    /* Japanese Kanji has no "Pin" ? so I used Yin :) */
								"", "", "",
                                "", "",  
							    "Ciyin",  /* Unicode */
							  };

	char *ZhiyinKeyboardMapStr[] = {
	     	"Auto Keyboard (No Tone)", 
		    "Pinyin Keyboard (No Tone)",
			"ShuangPin Keyboard (No Tone)",
		    "Standard Phone Keyboard (No Tone)", /* also called GuoQiao */
		    "Eten Phone Keyboard (No Tone)",
		    "ChingYeah Phone Keyboard(No Tone)",
		    "IBM Phone Keyboard(No Tone)",
			"ShenTong Phone Keyboard(No Tone)",
			"Eten26 Phone Keyboard (No Tone)",
			"Hsu25 Phone Keyboard (No Tone)",
	};

	char *CiyinKeyboardMapStr[] = {
	     	"Auto Keyboard (Optional Tone)", 
			"Pinyin Keyboard (Optional Tone -/=\\`)",
			"ShuangPin Keyboard(Optional Tone -/=\\`)",
		    "Standard Phone Keyboard (Optional Tone)",
		    "Eten Phone Keyboard (Optional Tone)",
		    "ChingYeah Phone Keyboard (Optional Tone)",
		    "IBM Phone Keyboard (Optional Tone)",
			"ShenTong Phone Keyboard(Optional Tone)",
			"Eten26 Phone Keyboard (Optional Tone)",
			"Hsu25 Phone Keyboard (Optional Tone)",
	};

Punct1 Punct1Str[] = {
	{'.', { "", "", "B", "C", "", "", "", "", } }, 
	{',', { "", "", "C",  "A", "", "", "", "", } },
    {';', { "", "", "G", "Q", "", "", "", "", } },
    {':', { "", "", "F", "G", "", "", "", "", } },
    {'?', { "", "", "H", "H", "", "", "", "", } },
    {'(', { "", "", "i", "]", "", "", "", "", } },
    {')', { "", "", "j", "^", "", "", "", "", } },
    {'^', { "ġ", "ġ", "cc", "KK", "", "", "", "", } },
    {'_', { "", "", "\x81\x5c\x81\x5c", "ww", "", "", "", "",  } },
   {'\\', { "", "", "A", "B", "", "", "", "", } },
    {'@', { "", "", "E", "O", "", "", "", "", } },
    {'&', { "", "", "\x81\x5c", "w", "", "", "", "", } },  
    {'$', { "", "", "", "D", "", "", "", "", } },  /* Big5 0xA244 is D, 0xA243 isC */
	{'!', { "", "", "I", "I", "", "", "", "", } },
	{ 0,  }, /* dummy */
 };

Punct2 Punct2Str[] = {  /* KSC has no this punct? */
    {'\"', 0, { { "", "", "g", "", "", "", "", "", },
			  	{ "", "", "h", "", "", "", "", "", } }  },
    {'\'', 0, { { "", "", "e",  "", "", "", "", "", } ,
				{ "", "", "f", "", "", "", "", "",  } }  },
	{ 0, 0, }  /* dummy */
};

Punct3 Punct3Str[] = {  
    { { '<', '>' }, 0, { { { "", "", "s", "m", "", "", "", "", },     /* << */
                           { "", "", "q", "q", "", "", "", "", } },        
                         { { "", "", "t", "n", "","", "", "",  },     /* >> */
                           { "", "", "r", "r", "", "", "", "", } } }
    },
    { { 0, 0 } },
};

