/* vi: set sw=4 ts=4: */
/*
 * CCE - Console Chinese Environment -
 * Copyright (C) 1998-2003 Rui He (rhe@3eti.com)
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE TERRENCE R. LAMBERT BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <fcntl.h>
#include <sys/types.h>

#include "config.h"

#ifdef HAVE_SYS_MMAN_H
#include <sys/mman.h>
#endif

#include "defs.h"

void *mmap_file(void *start, size_t len, int prot, int flags, int fd, off_t offset)
{
	void *ptr;
	int bytes;

#ifdef DEBUG
	fprintf(stderr, "mmap_file: start=0x%x len=%d fd=%d offset=0x%x\n", start, len, fd, offset);
#endif
	
	if (fd < 0 || offset < 0 || len <= 0 || len > 64*1024*1024 /* max 64MB? */
		||	lseek(fd, offset, SEEK_SET) == -1)
	{
		fprintf(stderr, "mmap_file: lseek failed?\n");		
		return (void *)-1;
	}
	
	ptr = malloc(len);
	if (ptr == NULL) 
	{
#ifdef DEBUG
		fprintf(stderr, "mmap_file: failed to malloc %d bytes.\n", len);
#endif
		return (void *)-1;
	}
	
	bytes = read(fd, ptr, len);
	if (bytes != len)
	{
#ifdef DEBUG
		fprintf(stderr, "mmap_file: read %d bytes, not %d\n", bytes, len);
#endif
		free(ptr);
		return (void *)-1;
	}
	
	return ptr;
}

int munmap_file(void *start, size_t len)
{
	if (start) free(start);
	return 0; /* always success */
}
